﻿import os
import time
import torch
import torch.nn as nn
from torchvision.utils import save_image
import matplotlib.pyplot as plt
import numpy as np
start = time.time()
if not os.path.exists('samples') : os.makedirs('samples')                       # フォルダ作成

class Discriminator(nn.Module):                                                 # 識別器(Discriminator)
    def __init__(self):
        super().__init__()
        self.model = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size=3, padding=1),                         # [batch_size, 16, 28, 28]
            nn.BatchNorm2d(16),nn.LeakyReLU(0.2, inplace=True),
            
            nn.AvgPool2d(2),
            nn.Conv2d(16, 32, kernel_size=3, padding=1),                        # [batch_size, 32, 14, 14]
            nn.BatchNorm2d(32),nn.LeakyReLU(0.2, inplace=True),            
        
            nn.AvgPool2d(2),
            nn.Conv2d(32, 64, kernel_size=3, padding=1),                        # [batch_size, 64, 7, 7]
            nn.BatchNorm2d(64),nn.LeakyReLU(0.2, inplace=True),
            
            nn.AvgPool2d(7),
            nn.Conv2d(64, 1, 1))                                                # [batch_size]
        
    def forward(self, x):return self.model(x).squeeze()                         # 識別結果(0～1)を出力,[batch_size]

class Generator(nn.Module):                                                     # 生成器(Generator)
    def __init__(self):
        super().__init__()
        self.net = nn.Sequential(
            nn.ConvTranspose2d(64, 64, 4, 1, 0),                                # [batch_size, 64, 4, 4]
            nn.BatchNorm2d(64),nn.ReLU(inplace=True),
        
            nn.ConvTranspose2d(64, 32, 4, 1, 0),                                # [batch_size, 32, 7, 7]
            nn.BatchNorm2d(32),nn.ReLU(inplace=True),
            
            nn.ConvTranspose2d(32, 16, 2, 2, 0),                                # [batch_size, 16, 14, 14]
            nn.BatchNorm2d(16),nn.ReLU(inplace=True),
            
            nn.ConvTranspose2d(16, 1, 2, 2, 0),                                 # [batch_size, 1, 28, 28]
            nn.Tanh())
            
    def forward(self, x):return self.net(x)                                     # 偽画像結果を出力, [batch_size, 1, 28, 28]

# パラメータ設定
num_epochs               = 1
batch_size, latent_size  = 32 , 64
data_set                 = 60000/batch_size 
    
G, D                     = Generator(), Discriminator()
G, D                     = nn.DataParallel(G), nn.DataParallel(D)
d_optimizer              = torch.optim.Adam(D.parameters(),lr=0.0002, betas=(0.5, 0.999))
g_optimizer              = torch.optim.Adam(G.parameters(),lr=0.0002, betas=(0.5, 0.999))               
loss_func                = nn.BCEWithLogitsLoss()
noise                    = torch.randn(batch_size, latent_size, 1, 1)                
real_labels              = torch.ones(batch_size)                               # 正解ラベル
fake_labels              = torch.zeros(batch_size)                              # 不正解ラベル
d_losses, g_losses       = np.zeros(num_epochs), np.zeros(num_epochs) 
real_scores, fake_scores = np.zeros(num_epochs), np.zeros(num_epochs)

train_data               = np.loadtxt('mnist_train.txt')                        # MNIST画像データ(60,000)読み込み
train_data               = (train_data.astype(np.float32) - 127.5) / 127.5      # 正規化
train_data               = train_data.reshape(int(data_set),-1)                 # 1875セットに分割
train_data               = torch.from_numpy(train_data)
    
# 学習開始
for epoch in range(num_epochs):
    for i, real_img in enumerate(train_data):                                   # 1セットのtrain_data:[batch_size, 1, 28, 28]
      # Discriminatorの訓練
        fake_img       = G(torch.randn(batch_size, latent_size, 1, 1))          # [batch_size, 64, 1, 1]から[batch_size, 1, 28, 28]の偽画像を作成     
        fake_score     = D(fake_img)                                            # [batch_size, 1, 28, 28]の偽画像の識別結果(0～1)を出力,[batch_size]
        d_loss_fake    = loss_func(fake_score, fake_labels)

        real_score     = D(real_img.reshape(batch_size, 1, 28, 28))
        d_loss_real    = loss_func(real_score, real_labels)
        
        d_loss         = d_loss_real + d_loss_fake                              # 実画像と偽画像のロスを合計
        d_optimizer.zero_grad()                                                 # 重み更新
        g_optimizer.zero_grad()
        d_loss.backward()
        d_optimizer.step()

      # generatorの学習
        fake_img         = G(torch.randn(batch_size, latent_size, 1, 1)) 
        fake_score       = D(fake_img)                                          # 偽画像の識別結果(0～1)を出力
        g_loss           = loss_func(fake_score, real_labels)
        d_optimizer.zero_grad()                                                 # 重み更新
        g_optimizer.zero_grad()
        g_loss.backward()
        g_optimizer.step()

      # 画像出力
        if (i+1) % 200 == 0:
            print('Epoch:{}, Step:{}, accuracy:{:.3f}'.format(epoch, i+1, fake_score.mean().item()))
                        
            with torch.no_grad():                                               # 固定のノイズを入れgeneratorの性能を確認する
                sample_img = G(noise)
                sample_img = sample_img.reshape(batch_size, 1, 28, 28)
                sample_img = ((sample_img + 1)/2).clamp(0,1)                    # データを0～1の値に変換
                save_image(sample_img, os.path.join('samples', 'f_img_epoch{}_{}.png'.format(epoch,i+1)))

      # データ分析
        d_losses[epoch]    += d_loss.item()/data_set
        g_losses[epoch]    += g_loss.item()/data_set
        real_scores[epoch] += real_score.mean().item()/data_set
        fake_scores[epoch] += fake_score.mean().item()/data_set

# データ出力
plt.figure()
plt.subplot(2,1,1)
plt.plot(range(1, num_epochs + 1), d_losses, label='d_loss')
plt.plot(range(1, num_epochs + 1), g_losses, label='g_loss')    
plt.legend()

plt.subplot(2,1,2)
plt.ylim(0, 1)
plt.plot(range(1, num_epochs + 1), fake_scores, label='fake_score')
plt.plot(range(1, num_epochs + 1), real_scores, label='real_score')
plt.legend()

plt.savefig('loss_accuracy.pdf')
plt.close()

real_img = real_img.reshape(batch_size, 1, 28, 28)
save_image(real_img, os.path.join('samples', 'sample_real_img.png'))             # 実画像のサンプル
print("time:" , time.time() - start)
